from six.moves.urllib_parse import urljoin
from datetime import datetime
from traitlets import HasTraits, Unicode, Int, Instance
from .forms import Response, FormEntry, DOMEntry
from .base import Base, _SERVICE_LEVELS, _PRIVACY_LEVELS


class NotebookMetadataConfig(HasTraits):
    '''Paperboy configuration object representing a Notebook (metadata component)'''
    username = Unicode()
    userid = Unicode()

    notebook = Unicode()
    privacy = Unicode()
    level = Unicode()
    requirements = Unicode(allow_none=True, default_value='')
    dockerfile = Unicode(allow_none=True, default_value='')

    jobs = Int()
    reports = Int()

    created = Instance(datetime)
    modified = Instance(datetime)

    def to_json(self, include_notebook=False):
        '''Convert NotebookMetadata to a JSON'''
        ret = {}
        ret['username'] = self.username
        # ret['userid'] = self.userid
        if include_notebook:
            ret['notebook'] = self.notebook
        ret['privacy'] = self.privacy
        ret['level'] = self.level
        ret['jobs'] = self.jobs
        ret['reports'] = self.reports
        ret['created'] = self.created.strftime('%m/%d/%Y %H:%M:%S')
        ret['modified'] = self.modified.strftime('%m/%d/%Y %H:%M:%S')
        return ret

    @staticmethod
    def from_json(jsn):
        '''Create NotebookMetadata from a JSON'''
        ret = NotebookMetadataConfig()
        for k, v in jsn.items():
            if k in ('created', 'modified'):
                ret.set_trait(k, datetime.strptime(v, '%m/%d/%Y %H:%M:%S'))
            else:
                ret.set_trait(k, v)
        return ret


class NotebookConfig(Base):
    '''Paperboy configuration object representing a Notebook'''
    name = Unicode()
    id = Unicode()
    meta = Instance(NotebookMetadataConfig)

    def to_json(self, include_notebook=False):
        '''Convert Notebook to a JSON'''
        ret = {}
        ret['name'] = self.name
        ret['id'] = self.id
        ret['meta'] = self.meta.to_json(include_notebook)
        return ret

    def form(self):
        '''Generate Form template for client from a Notebook object'''
        f = Response()
        f.entries = [
            FormEntry(name='file', type='file', label='File', required=True),
            FormEntry(name='name', type='text', label='Name', placeholder='Name for Notebook...', required=True),
            FormEntry(name='privacy', type='select', label='Privacy', options=_PRIVACY_LEVELS, required=True),
            FormEntry(name='level', type='select', label='level', options=_SERVICE_LEVELS, required=True),
            FormEntry(name='build', type='label', label='Build options'),
            FormEntry(name='requirements', type='file', label='requirements.txt', required=False),
            FormEntry(name='dockerfile', type='file', label='Dockerfile', required=False),
            FormEntry(name='submit', type='submit', value='save', url=urljoin(self.config.apiurl, 'notebooks?action=save')),
        ]
        return f.to_json()

    @staticmethod
    def from_json(jsn, config):
        '''Create Notebook from a JSON'''
        ret = NotebookConfig(config)
        ret.name = jsn.pop('name')
        ret.id = jsn.pop('id')

        if 'meta' in jsn:
            ret.meta = NotebookMetadataConfig.from_json(jsn['meta'])
        else:
            ret.meta = NotebookMetadataConfig.from_json(jsn)

        return ret

    def edit(self):
        '''Generate Edit template for client from a Notebook object'''
        f = Response()
        f.entries = [
            FormEntry(name='name', type='text', value=self.name, placeholder='Name for Job...', required=True),
            FormEntry(name='id', type='text', value=self.id, label='Id', hidden=True),
            FormEntry(name='privacy', type='select', value=self.meta.privacy, label='Visibility', options=_PRIVACY_LEVELS, required=True),
            FormEntry(name='level', type='select', value=self.meta.level, label='Level', options=_SERVICE_LEVELS, required=True),
            FormEntry(name='notebook', type='json', value=self.meta.notebook, placeholder='Notebook json...', required=True)
        ]

        f.entries.append(FormEntry(name='requirements', type='textarea', value=self.meta.requirements, label='requirements.txt', required=False))
        f.entries.append(FormEntry(name='dockerfile', type='textarea', value=self.meta.dockerfile, label='Dockerfile', required=False))

        f.entries.append(FormEntry(name='save', type='submit', value='save', url=urljoin(self.config.apiurl, 'notebooks?action=save')))
        f.entries.append(FormEntry(name='delete', type='submit', value='delete', url=urljoin(self.config.apiurl, 'notebooks?action=delete')))
        return f.to_json()

    def entry(self):
        '''Generate ListTable entry for client from a Notebook object'''
        f = Response()
        f.entries = [
            DOMEntry(name='name', type='label', value=self.name, label='Name'),
            DOMEntry(name='id', type='label', value=self.id, label='Id', hidden=True),
            DOMEntry(name='visibility', type='label', value=self.meta.privacy, label='Visibility'),
            DOMEntry(name='level', type='label', value=self.meta.level, label='Level'),
            DOMEntry(name='notebook', type='ipynb', value=self.meta.notebook, label='Notebook'),
            DOMEntry(name='requirements', type='textfile', value=self.meta.requirements, label='requirements'),
            DOMEntry(name='dockerfile', type='textfile', value=self.meta.dockerfile, label='Dockerfile'),
            DOMEntry(name='jobs', type='label', value=str(self.meta.jobs), label='Jobs'),
            DOMEntry(name='reports', type='label', value=str(self.meta.reports), label='Reports'),
            DOMEntry(name='created', type='label', value=self.meta.created.strftime('%m/%d/%Y %H:%M:%S'), label='Created'),
            DOMEntry(name='modified', type='label', value=self.meta.modified.strftime('%m/%d/%Y %H:%M:%S'), label='Modified'),
            DOMEntry(name='delete', type='button', value='delete', label='delete', url=urljoin(self.config.apiurl, 'notebooks?action=delete'))
        ]
        return f.to_json()

    def store(self):
        '''Generate response modal for client when saving a Notebook object'''
        ret = Response()
        ret.entries = [
            DOMEntry(type='p', value='Success!'),
            DOMEntry(type='p', value='Successfully stored notebook: {}'.format(self.name)),
        ]
        return ret.to_json()



from sqlalchemy import Column, Integer, String
from sqlalchemy.orm import relationship
from .base import Base


class UserSQL(Base):
    __tablename__ = 'users'
    id = Column(Integer, primary_key=True)
    name = Column(String)
    password = Column(String)

    notebooks = relationship('NotebookSQL', back_populates='user')
    jobs = relationship('JobSQL', back_populates='user')
    reports = relationship('ReportSQL', back_populates='user')

    def __repr__(self):
        return "<User(name='%s')>" % self.name



from .common import BaseTask
import json
import luigi
from luigi.parameter import Parameter, ParameterVisibility


class NBConvertTask(BaseTask):
    report = Parameter(visibility=ParameterVisibility.HIDDEN)

    def __init__(self, *args, **kwargs):
        super(NBConvertTask, self).__init__(*args, **kwargs)
        self._report = json.loads(self.report)

    def run(self):
        self.log.critical('nbconvert')

        fp = self.input().open('r')
        papermilled = fp.read()
        fp.close()

        if self._report['meta']['output'] != 'notebook':
            from paperboy.worker import run_nbconvert

            template = self._report['meta'].get('template', '')

            ret = run_nbconvert(self._report['meta']['notebook'],
                                papermilled,
                                self._report['meta']['output'],
                                template,
                                self._report['meta']['strip_code'],
                                )
        else:
            ret = papermilled

        fp = self.output().open('w')
        fp.write(ret.decode('utf8'))
        fp.close()
        self._completed = True

    def output(self):
        return luigi.LocalTarget(self._report['name'])



import os
import os.path
import subprocess
import sys
try:
    from tempfile import TemporaryDirectory
except ImportError:
    from backports.tempfile import TemporaryDirectory


def run(nb_name, nb_text, to='html', template='', hide_input=False):
    '''Helper function to run nbconvert, used by airflow

    Args:
        nb_name (string): Name of notebook
        nb_text (string): nbformat json of text of notebook to convert
        to (string): what to convert to, options dictated by nbconvert
        template (string): path to template to use
        hide_input (boolean): hide code
    Returns:
        string: text/binary output of nbconvert
    '''

    with TemporaryDirectory() as tempdir:
        in_file = os.path.join(tempdir, '{}.ipynb'.format(nb_name))
        out_file = os.path.join(tempdir, '{}_out'.format(nb_name))

        # hack for pdfs, .pdf appended automatically by xelatex
        if to != 'pdf':
            out_file += '.' + to

        with open(in_file, 'w') as fp:
            fp.write(nb_text)

        # assemble nbconvert command
        argv = []
        argv = [sys.executable, '-m', 'nbconvert', '--to', to]

        # pass in template arg
        if template:
            argv.extend(['--template', template])
        if hide_input:
            argv.append('--no-input')
        argv.append('--no-prompt')

        # output to outname
        argv.extend([in_file, '--output', out_file])

        subprocess.call(argv)

        # hack for pdfs
        if to == 'pdf':
            out_file += '.' + to

        # fail if doesnt exist
        with open(out_file, 'rb') as fp:
            ret = fp.read()
    return ret



import falcon
import json
from .base import BaseResource


class ConfigResource(BaseResource):
    '''Falcon resource to get form entries'''
    def __init__(self, *args, **kwargs):
        super(ConfigResource, self).__init__(*args, **kwargs)

    def on_get(self, req, resp):
        '''Get configuration page to create a new notebook/job/report'''
        resp.content_type = 'application/json'
        type = req.params.get('type', None)
        if type is None:
            resp.body = json.dumps(self.config.to_dict())
        elif type == 'notebooks':
            resp.body = json.dumps(self.db.notebooks.form())
        elif type == 'jobs':
            resp.body = json.dumps(self.db.jobs.form())
        elif type == 'reports':
            resp.body = json.dumps(self.db.reports.form())
        else:
            resp.status = falcon.HTTP_404



from traitlets import List, Int, Unicode, Bool, Instance, HasTraits, validate, TraitError
from .base import Base

_DOM_IMPLEMENTED = ('text', 'select', 'label', 'button', 'p', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'p', 'span', 'json', 'ipynb', 'textfile')
_FORM_IMPLEMENTED = ('file', 'text', 'select', 'label', 'submit', 'datetime', 'autocomplete', 'checkbox', 'textarea', 'json')


class FormEntry(HasTraits):
    '''Form template entry to be rendered on the client'''
    name = Unicode(allow_none=False)
    type = Unicode(default_value='text')

    @validate('type')
    def _validate_type(self, proposal):
        if proposal['value'] not in _FORM_IMPLEMENTED:
            raise TraitError('Unrecognized type : {}'.format(proposal['value']))
        return proposal['value']

    value = Unicode(default_value='')
    label = Unicode(allow_none=True)
    placeholder = Unicode(allow_none=True)
    options = List(default_value=[])
    required = Bool(default_value=False)
    readonly = Bool(default_value=False)
    url = Unicode(allow_none=True)
    hidden = Bool(default_value=False)

    def to_json(self):
        '''Convert form entry to JSON'''
        ret = {}
        ret['name'] = self.name
        ret['type'] = self.type
        if self.value:
            ret['value'] = self.value
        if self.label:
            ret['label'] = self.label
        if self.placeholder:
            ret['placeholder'] = self.placeholder
        if self.options:
            ret['options'] = self.options
        if self.url:
            ret['url'] = self.url

        ret['required'] = self.required
        ret['readonly'] = self.readonly
        ret['hidden'] = self.hidden

        return ret


class DOMEntry(HasTraits):
    '''DOM node template to be rendered on the client'''
    name = Unicode(allow_none=False)
    type = Unicode(default_value='p')

    @validate('type')
    def _validate_type(self, proposal):
        if proposal['value'] not in _DOM_IMPLEMENTED:
            raise TraitError('Unrecognized type : {}'.format(proposal['value']))
        return proposal['value']

    value = Unicode(default_value='')
    label = Unicode(allow_none=True)
    placeholder = Unicode(allow_none=True)
    options = List(default_value=[])
    required = Bool(default_value=False)
    readonly = Bool(default_value=False)
    hidden = Bool(default_value=False)

    def to_json(self):
        ret = {}
        ret['name'] = self.name
        ret['type'] = self.type
        if self.value:
            ret['value'] = self.value
        if self.label:
            ret['label'] = self.label
        if self.placeholder:
            ret['placeholder'] = self.placeholder
        if self.options:
            ret['options'] = self.options
        ret['required'] = self.required
        ret['readonly'] = self.readonly
        ret['hidden'] = self.hidden
        return ret

    @staticmethod
    def from_json(jsn):
        ret = DOMEntry()

        for k, v in jsn.items():
            ret.set_trait(k, v)
        return ret


class Response(HasTraits):
    '''Response modal template for client'''
    entries = List()

    def to_json(self):
        ret = []
        for entry in self.entries:
            ret.append(entry.to_json())
        return ret


class ListResult(HasTraits):
    '''List result metadata for pagination'''
    page = Int(default_value=1)
    pages = Int(default_value=1)
    count = Int(default_value=1)
    total = Int(default_value=1)
    results = List(trait=Instance(Base))

    def to_json(self):
        ret = {}
        ret['page'] = self.page
        ret['pages'] = self.pages
        ret['count'] = self.count
        ret['total'] = self.total
        ret['results'] = [r.entry() for r in self.results]
        return ret



# -*- coding: utf-8 -*-
import configparser
import json
import os
import os.path
import jinja2
import subprocess
from base64 import b64encode
from sqlalchemy import create_engine
from ..base import BaseScheduler, interval_to_cron

with open(os.path.abspath(os.path.join(os.path.dirname(__file__), 'paperboy.airflow.py')), 'r') as fp:
    TEMPLATE = fp.read()

QUERY = '''
SELECT task_id, dag_id, execution_date, state, unixname
FROM task_instance
ORDER BY execution_date ASC
LIMIT 20;
'''


class AirflowScheduler(BaseScheduler):
    def __init__(self, *args, **kwargs):
        '''Create a new airflow scheduler, connecting to the airflow instances configuration'''
        super(AirflowScheduler, self).__init__(*args, **kwargs)
        cp = configparser.ConfigParser()
        cp.read(self.config.scheduler_config.config)
        try:
            self.sql_conn = cp['core']['sql_alchemy_conn']
        except KeyError:
            self.sql_conn = ''

        if self.sql_conn:
            self.engine = create_engine(self.sql_conn)
        else:
            self.engine = None

    def status(self, user, params, session, *args, **kwargs):
        '''Get status of job/report DAGs'''
        type = params.get('type', '')
        gen = AirflowScheduler.query(self.engine)
        if type == 'jobs':
            return gen['jobs']
        elif type == 'reports':
            return gen['reports']
        else:
            return gen

    @staticmethod
    def query(engine):
        '''Get status of job/report DAGs from airflow\'s database'''
        ret = {'jobs': [], 'reports': []}
        with engine.begin() as conn:
            res = conn.execute(QUERY)
            for i, item in enumerate(res):
                ret['jobs'].append(
                    {'name': item[1],
                     'id': item[1][4:],
                     'meta': {
                        'id':  item[1][4:],
                        'execution': item[2].strftime('%m/%d/%Y %H:%M:%S'),
                        'status': '✔' if item[3] == 'success' else '✘'}
                     }
                )

                report_name = item[0].replace('ReportPost-', '') \
                                     .replace('Report-', '') \
                                     .replace('ReportNBConvert-', '') \
                                     .replace('ReportPapermill-', '')

                report_type = 'Post' if 'ReportPost' in item[0] else \
                              'Papermill' if 'Papermill' in item[0] else \
                              'NBConvert' if 'NBConvert' in item[0] else \
                              'Setup'

                ret['reports'].append(
                    {'name': item[0],
                     'id': report_name,
                     'meta': {
                        'run': item[2].strftime('%m/%d/%Y %H:%M:%S'),
                        'status': '✔' if item[3] == 'success' else '✘',
                        'type': report_type
                        }
                     }
                )
            return ret

    @staticmethod
    def template(config, user, notebook, job, reports, *args, **kwargs):
        '''jinja templatize airflow DAG for paperboy (paperboy.airflow.py)'''
        owner = user.name
        start_date = job.meta.start_time.strftime('%m/%d/%Y %H:%M:%S')
        email = 'test@test.com'
        job_json = b64encode(json.dumps(job.to_json(True)).encode('utf-8'))
        report_json = b64encode(json.dumps([r.to_json() for r in reports]).encode('utf-8'))
        interval = interval_to_cron(job.meta.interval, job.meta.start_time)

        tpl = jinja2.Template(TEMPLATE).render(
            owner=owner,
            start_date=start_date,
            interval=interval,
            email=email,
            job_json=job_json,
            report_json=report_json,
            output_config=json.dumps(config.output.to_json())
            )
        return tpl

    def schedule(self, user, notebook, job, reports, *args, **kwargs):
        '''Schedule a DAG for `job` composed of `reports` to be run on airflow'''
        template = AirflowScheduler.template(self.config, user, notebook, job, reports, *args, **kwargs)
        name = job.id + '.py'
        os.makedirs(self.config.scheduler_config.dagbag, exist_ok=True)
        with open(os.path.join(self.config.scheduler_config.dagbag, name), 'w') as fp:
            fp.write(template)
        return template

    def unschedule(self, user, notebook, job, reports, *args, **kwargs):
        '''Remove the DAG for `user` and `notebook` composed of `job` running `reports` from
            airflow 2 parts, remove the dag from disk and delete the dag from airflow's database using the CLI'''
        if reports:
            # reschedule
            return self.schedule(user, notebook, job, reports, *args, **kwargs)

        else:
            # delete
            name = job.id + '.py'
            file = os.path.join(self.config.scheduler_config.dagbag, name)
            dag = 'DAG-' + job.id

            # delete dag file
            os.remove(file)

            # delete dag
            # FIXME
            try:
                cmd = ['airflow', 'delete_dag', dag, '-y']
                subprocess.call(cmd)
            except Exception as e:
                print(e)



import os
import getpass


def schedule_cron(command, interval, crontab=None):
    from crontab import CronTab, CronItem
    if not os.path.exists(crontab):
        with open(crontab, 'w'):
            pass

    if crontab:
        c = CronTab(tabfile=crontab)
    else:
        c = CronTab(user=getpass.getuser())

    job = CronItem.from_line(interval + ' ' + command, cron=c)

    c.append(job)
    c.write()
    return c, job


def unschedule_cron(command, crontab=None):
    from crontab import CronTab
    if crontab:
        c = CronTab(tabfile=crontab)
    else:
        c = CronTab(user=getpass.getuser())

    for line in c.lines:
        if str(line) == command:
            line.delete()
            c.write()
            return True
    return False



# -*- coding: utf-8 -*-
#
# FIXME: this code should be used to configure airflow
# as the login authority for paperboy. It is not yet
# complete.
#
#
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
"""
Override this file to handle your authenticating / login.

Copy and alter this file and put in your PYTHONPATH as airflow_login.py,
the new module will override this one.
"""

import flask_login
# from flask_login import login_required, current_user, logout_user
from flask import url_for, redirect
from airflow import settings
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from paperboy.storage.sqla.models.user import UserSQL


DEFAULT_USERNAME = 'anon'

login_manager = flask_login.LoginManager()
login_manager.login_view = 'airflow.login'  # Calls login() below
login_manager.login_message = None


class DefaultUser(object):
    def __init__(self, user):
        self.user = user

    @property
    def is_active(self):
        """Required by flask_login"""
        return False

    @property
    def is_authenticated(self):
        """Required by flask_login"""
        return False

    @property
    def is_anonymous(self):
        """Required by flask_login"""
        return True

    def data_profiling(self):
        """Provides access to data profiling tools"""
        return False

    def is_superuser(self):
        """Access all the things"""
        return False


@login_manager.user_loader
def load_user(userid):
    engine = create_engine(settings.webserver.paperboy_sql, echo=False)
    sm = sessionmaker(bind=engine)

    with sm() as session:
        user = session.query(UserSQL).filter(UserSQL.id == userid).first()
        return DefaultUser(user)


def login(self, request):
    engine = create_engine(settings.webserver.paperboy_sql, echo=False)
    sm = sessionmaker(bind=engine)

    with sm() as session:
        user = session.query(UserSQL).filter(UserSQL.name == DEFAULT_USERNAME).first()

        if not user:
            user = UserSQL(name=DEFAULT_USERNAME)
        session.merge(user)
        session.commit()
        flask_login.login_user(DefaultUser(user))
        session.commit()
    return redirect(request.args.get("next") or url_for("index"))



import logging
import math
from sqlalchemy import or_
from paperboy.config import ListResult


class BaseSQLStorageMixin(object):
    '''Base class to share common SQL operations between User/Notebook/Job/Report backends'''

    def _form(self, ConfigCls):
        '''Generate form from configuration type.

        Args:
            ConfigCls (paperboy.config.base.Base): UserConfig/NotebookConfig/JobConfig/ReportConfig
        Returns:
            paperboy.config.forms.Response
        '''
        return ConfigCls(self.config).form()

    def _search(self, SqlCls, ClsName, user, params, session, *args, **kwargs):
        '''Search for a given name in instances of ClsName stored in SQL as type SqlCls

        Args:
            SqlCls (BaseSQLStorageMixin + BaseStorage): Class implementing the BaseStorage interface
            ClsName (string): Notebook/Job/Report
            user (paperboy.storage.base.UserStorage): the user access this data
            params (dict): query parameters
            session (sqlalchemy session): the active sqlalchemy session
        Returns:
            list: results with name/id
        '''
        name = params.get('name', '')
        count = int(params.get('count', 10))
        if name is None or name == '':
            return []
        nbs = session.query(SqlCls) \
            .filter(SqlCls.name.like(lookfor(name))) \
            .filter(SqlCls.userId == int(user.id) or
                    (hasattr(SqlCls, 'privacy') and SqlCls.privacy == 'public')) \
            .limit(count)

        return [{'id': ClsName + '-' + str(nb.id), 'name': nb.name} for nb in nbs]

    def _list(self, SqlCls, setter, user, params, session, *args, **kwargs):
        '''List all instances of a give storage class

        Args:
            SqlCls (BaseSQLStorageMixin + BaseStorage): Class implementing the BaseStorage interface
            setter (string): unused
            user (paperboy.storage.base.UserStorage): the user access this data
            params (dict): query parameters
            session (sqlalchemy session): the active sqlalchemy session
        Returns:
            list of paperboy.config.base.Base: result notebooks/jobs/reports as json
        '''
        base = session.query(SqlCls) \
            .filter(or_(SqlCls.userId.like((user.id)),
                        (hasattr(SqlCls, 'privacy') and SqlCls.privacy == 'public'))) \

        page = int(params.get('page', 1))
        nbs = base[25*(page-1):25*page]

        result = ListResult()
        result.total = base.count()
        result.count = len(nbs)
        result.page = page
        result.pages = math.ceil(result.total/25) if result.count > 0 else 1

        logging.critical('list : {}, result : {} - {}'.format(SqlCls, result.total, len(nbs)))

        result.results = [x.to_config(self.config) for x in nbs]
        return result.to_json()

    def _detail(self, SqlCls, user, params, session, *args, **kwargs):
        '''Get detailed view of instance

        Args:
            SqlCls (BaseSQLStorageMixin + BaseStorage): Class implementing the BaseStorage interface
            user (paperboy.storage.base.UserStorage): the user access this data
            params (dict): query parameters
            session (sqlalchemy session): the active sqlalchemy session
        Returns:
            edit of paperboy.config.base.Base: notebook/job/report config edit
        '''
        id = justid(params.get('id') or -1)
        try:
            id = int(id)
        except ValueError:
            return {}

        if id < 0:
            return {}

        nb_sql = session.query(SqlCls).get(id)  # FIXME permission?

        logging.critical('detail : {}, result : {}'.format(id, nb_sql))

        if nb_sql:
            return nb_sql.to_config(self.config).edit()
        return {}


def lookfor(s):
    '''Helpful matcher for sqlalchemy query'''
    if '*' in s or '_' in s:
        return s.replace('_', '__')\
                .replace('*', '%')\
                .replace('?', '_')
    return '%{0}%'.format(s)


def justid(id):
    '''Helper method to turn name-id string into just id'''
    if isinstance(id, int):
        return id
    if '-' in id:
        return id.split('-', 1)[-1]
    return id



from datetime import datetime
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from .models.base import Base
from .models.user import UserSQL
from .models.notebook import NotebookSQL
from .models.job import JobSQL
from .models.report import ReportSQL


def main(sql_url):
    '''Create dummy notebook data for sqlalchemy'''
    engine = create_engine(sql_url, echo=False)
    Base.metadata.create_all(engine)
    sm = sessionmaker(bind=engine)

    session = sm()
    user = UserSQL(name='test')
    session.add(user)
    session.commit()
    session.refresh(user)

    notebook = '''
    '''
    created = datetime.now()
    modified = datetime.now()

    notebook = NotebookSQL(name='MyNotebook',
                           userId=user.id,
                           user=user,
                           notebook='',
                           privacy='',
                           level='',
                           requirements='',
                           dockerfile='',
                           created=created,
                           modified=modified)
    session.add(notebook)
    session.commit()
    session.refresh(notebook)

    job = JobSQL(name='MyJob',
                 userId=user.id,
                 user=user,
                 notebookId=notebook.id,
                 notebook=notebook,
                 start_time=created,
                 interval='minutely',
                 level='production',
                 created=created,
                 modified=modified)
    session.add(job)
    session.commit()
    session.refresh(job)

    type = 'run'
    output = 'notebook'
    strip_code = True

    created = datetime.now()
    modified = datetime.now()

    name = job.name + '-Report-1'

    param = '{}'
    report = ReportSQL(name=name,
                       userId=user.id,
                       user=user,
                       notebookId=notebook.id,
                       notebook=notebook,
                       jobId=job.id,
                       job=job,
                       type=type,
                       output=output,
                       strip_code=strip_code,
                       parameters=param,
                       created=created,
                       modified=modified)
    session.add(report)
    session.commit()
    session.refresh(report)



import os
import os.path
import logging
from paperboy.utils import name_to_class
from ._nbconvert import run as run_nbconvert  # noqa: F401
from ._papermill import run as run_papermill  # noqa: F401


def run(job, reports, working_dir=None):
    logging.critical('Calling run on job - %s' % str(job.id))
    for rep in reports:
        # type is "convert" for nbconversion jobs and "publish" for publish jobs
        type = rep.meta.type

        # Papermill task, performs the report creation
        # using papermill and the report's individual
        # parameters and configuration
        papermilled = run_papermill(rep.meta.notebook.name,
                                    rep.meta.notebook.meta.notebook,
                                    rep.meta.parameters,
                                    rep.meta.strip_code)

        if type == 'convert':
            # NBConvert task, performs the NBConversion if
            # required
            if rep.meta.output != 'notebook':
                template = rep.meta.template

                ret = run_nbconvert(rep.meta.notebook.name,
                                    papermilled,
                                    rep.meta.output,
                                    template,
                                    rep.meta.strip_code,
                                    )
            else:
                ret = papermilled

        elif type == 'publish':
            # Dokku deployment task, performs the
            # deploy to the dokku repo
            raise NotImplementedError()

            # Assemble a Voila Job from papermilled notebook
        else:
            raise NotImplementedError()

        # The post-report task, used for post-report
        # tasks such as sending the report in an email
        # or pushing the deploy to dokku

        config = {"type": "local",
                  "dir": os.path.expanduser('~/Downloads'),
                  "clazz": "paperboy.output.local.LocalOutput",
                  "config": "paperboy.config.output.LocalOutputConfig"}
        config = name_to_class(config['config']).from_json(config)
        outputter = config.clazz(config)
        outputter.write(rep, ret)



# This is the jinja2 templatized set of tasks that
# generates the job. It has the following form as an
# example:
#
#    Job -> Report -> Papermill -> NBConvert ->      ReportPost -\
#      \--> Report -> Papermill -> Voila -> Dokku -> ReportPost --\
#       \-> Report -> Papermill -> NBConvert ->      ReportPost ----> Job Cleanup
#

import luigi
import os
import os.path
import json
from base64 import b64decode

# Base Job Tasks
from paperboy.scheduler.luigi.luigi_tasks import JobTask, JobCleanupTask

# Base Report Tasks
from paperboy.scheduler.luigi.luigi_tasks import ReportTask, ReportPostTask

# Convert Tasks
from paperboy.scheduler.luigi.luigi_tasks import PapermillTask, NBConvertTask

# Publish operators
from paperboy.scheduler.luigi.luigi_tasks import VoilaTask, DokkuTask

from datetime import datetime

###################################
# Inline job and reports as b64 json  #
###################################
job_string = b64decode({{job_json}})
job_json = json.loads(job_string)
reports_string = b64decode({{report_json}})
reports_json = json.loads(reports_string)

interval = "{{interval}}"
start_date = datetime.strptime('{{start_date}}', '%m/%d/%Y %H:%M:%S')
owner = "{{owner}}"
email = "{{email}}"


###################################
# Create dag from job and reports #
###################################
def build():
    # The Job task, used for bundling groups of reports,
    # setting up env/image
    job = JobTask(job=job_string,
                  task_id='Job-{}'.format(job_json['id']))

    cleanup_requires = []

    for rep in reports_json:
        # copy over notebook text (only store 1 copy in the job json)
        rep['meta']['notebook_text'] = job_json['meta']['notebook_text']

        # type is "convert" for nbconversion jobs and "publish" for publish jobs
        type = rep['meta']['type']

        # Report task, performs the required
        # steps prior to running the report
        r = ReportTask(report=json.dumps(rep),
                       task_id='Report-{}'.format(rep['id']))
        r._reqs = job

        # Papermill task, performs the report creation
        # using papermill and the report's individual
        # parameters and configuration
        pp = PapermillTask(report=json.dumps(rep),
                           task_id='ReportPapermill-{}'.format(rep['id']))
        pp._reqs = r

        if type == 'convert':
            # NBConvert task, performs the NBConversion if
            # required
            nb = NBConvertTask(report=json.dumps(rep),
                               task_id='ReportNBConvert-{}'.format(rep['id']))
            nb._reqs = pp

        elif type == 'publish':
            # Dokku deployment task, performs the
            # deploy to the dokku repo
            d = DokkuTask(report=json.dumps(rep),
                          task_id='ReportDokku-{}'.format(rep['id']))
            d._reqs = pp

            # Assemble a Voila Job from papermilled notebook
            nb = VoilaTask(report=json.dumps(rep),
                           task_id='ReportVoila-{}'.format(rep['id']))
            nb._reqs = d

        else:
            raise NotImplementedError()

        # The post-report task, used for post-report
        # tasks such as sending the report in an email
        # or pushing the deploy to dokku

        # FIXME
        rp = ReportPostTask(report=json.dumps(rep),
                            config='{"type": "local", "dir": "' +
                            os.path.expanduser("~/Downloads") +
                            '", "clazz": "paperboy.output.local.LocalOutput", "config": "paperboy.config.output.LocalOutputConfig"}',
                            task_id='ReportPost-{}'.format(rep['id']))
        rp._reqs = nb
        cleanup_requires.append(rp)

    # The cleanup task, run after all reports are finished
    cleanup = JobCleanupTask(job=job_string,
                             task_id='JobCleanup-{}'.format(job_json['id']),
                             owner=owner,
                             start_date=start_date,
                             interval=interval,
                             email=email)
    cleanup._reqs = cleanup_requires
    cleanup.time = datetime.now()
    return cleanup

if __name__ == '__main__':
    luigi.build([build()])



from .common import JobOperator, JobCleanupOperator, ReportOperator  # noqa: F401
from .dokku import DokkuOperator  # noqa: F401
from .nbconvert import NBConvertOperator  # noqa: F401
from .papermill import PapermillOperator  # noqa: F401
from .post import ReportPostOperator  # noqa: F401
from .voila import VoilaOperator  # noqa: F401



import falcon
import os
import os.path
import mimetypes
# from functools import lru_cache
from .base import BaseResource


# @lru_cache(20)
def read(file):
    '''read static resource from disk (js/css/etc)'''
    path = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'assets', 'static', file.replace('/static/', '')))
    if not os.path.exists(path):
        return None
    with open(path, 'rb') as fp:
        return fp.read()


class StaticResource(BaseResource):
    '''Falcon resource to service CSS/JS/image/font files'''

    auth_required = False

    def __init__(self, *args, **kwargs):
        super(StaticResource, self).__init__(*args, **kwargs)

    def on_get(self, req, resp):
        '''read and return resource with appropriate mimetype'''
        filetype = mimetypes.guess_type(req.path, strict=True)[0]
        resp.content_type = filetype
        file = read(req.path)

        if file:
            resp.body = file
        else:
            resp.status = falcon.HTTP_404



import os
import json
import logging
from datetime import datetime
from paperboy.config import ReportConfig
from paperboy.storage import ReportStorage
from .base import BaseSQLStorageMixin, justid
from .models.user import UserSQL
from .models.notebook import NotebookSQL
from .models.job import JobSQL
from .models.report import ReportSQL


class ReportSQLStorage(BaseSQLStorageMixin, ReportStorage):
    def status(self, user, params, session, *args, **kwargs):
        base = session.query(ReportSQL) \
            .filter(ReportSQL.userId == int(user.id))

        return {'total': base.count(),
                'notebook': base.filter(ReportSQL.output == 'notebook').count(),
                'pdf':  base.filter(ReportSQL.output == 'pdf').count(),
                'html': base.filter(ReportSQL.output == 'html').count(),
                'email': base.filter(ReportSQL.output == 'email').count(),
                'script': base.filter(ReportSQL.output == 'script').count()}

    def form(self):
        '''Pass through to shared method in BaseSQLStorageMixin'''
        return self._form(ReportConfig)

    def search(self, user, params, session, *args, **kwargs):
        '''Pass through to shared method in BaseSQLStorageMixin'''
        return self._search(ReportSQL, 'Report', user, params, session, *args, **kwargs)

    def list(self, user, params, session, *args, **kwargs):
        '''Pass through to shared method in BaseSQLStorageMixin'''
        return self._list(ReportSQL, 'reports', user, params, session, *args, **kwargs)

    def detail(self, user, params, session, *args, **kwargs):
        '''Pass through to shared method in BaseSQLStorageMixin'''
        return self._detail(ReportSQL, user, params, session, *args, **kwargs)

    def store(self, user, params, session, *args, **kwargs):
        name = params.get('name')
        user_sql = session.query(UserSQL).get(int(user.id))

        notebookid = params.get('notebook')
        nb_sql = session.query(NotebookSQL).get(int(justid(notebookid)))

        jobid = params.get('job')
        jb_sql = session.query(JobSQL).get(int(justid(jobid)))

        type = params.get('type') or 'run'
        output = params.get('output') or 'pdf'
        strip_code = params.get('strip_code') == 'yes'
        parameters = params.get('parameters') or '[]'

        created = datetime.now()
        modified = datetime.now()

        id = params.get('id')
        if id:
            id = justid(id)
            rp = session.query(ReportSQL).filter(ReportSQL.id == id).first()
            rp.name = name
            rp.type = type
            rp.output = output
            rp.strip_code = strip_code
            rp.parameters = parameters
            rp.modified = modified

        else:
            rp = ReportSQL(name=name,
                           userId=user.id,
                           user=user_sql,
                           notebookId=notebookid,
                           notebook=nb_sql,
                           jobId=jobid,
                           job=jb_sql,
                           type=type,
                           output=output,
                           strip_code=strip_code,
                           parameters=parameters,
                           created=created,
                           modified=modified)

        session.add(rp)

        # generate id
        session.flush()
        session.refresh(rp)

        store = rp.to_config(self.config).store()
        logging.critical("Storing report {}".format(rp))
        return store

    def generate(self, user, params, session, *args, **kwargs):
        autogen = params.get('autogen') or False
        ret = []

        if autogen:
            parameters_inline = params.get('parameters_inline') or ''
            parameters = params.get('parameters')

            if not parameters_inline and parameters:
                report_params = parameters.file.read().decode('utf-8')
            else:
                report_params = parameters_inline
            if not report_params:
                report_params = []
            else:
                report_params = [json.loads(p) for p in report_params.split(os.linesep)]

            user_sql = session.query(UserSQL).get(int(user.id))

            notebook = params.get('notebook')
            nb_sql = session.query(NotebookSQL).get(int(justid(notebook)))

            job_name = params.get('name')
            job = kwargs.get('job')
            job_id = int(justid(job.id))
            jb_sql = session.query(JobSQL).get(int(justid(job_id)))

            type = params.get('type') or 'run'
            output = params.get('output') or 'notebook'
            strip_code = params.get('strip_code') == 'yes'

            created = datetime.now()
            modified = datetime.now()

            for i, param in enumerate(report_params):
                name = job_name + '-Report-' + str(i)
                param = json.dumps(param)
                rp = ReportSQL(name=name,
                               userId=user.id,
                               user=user_sql,
                               notebookId=notebook,
                               notebook=nb_sql,
                               jobId=job_id,
                               job=jb_sql,
                               type=type,
                               output=output,
                               strip_code=strip_code,
                               parameters=param,
                               created=created,
                               modified=modified)
                session.add(rp)

                # FIXME
                session.flush()
                session.refresh(rp)
                ret.append(rp.to_config(self.config))

            session.flush()

        return ret

    def delete(self, user, params, session, *args, **kwargs):
        # TODO only if allowed
        id = justid(params.get('id'))
        rp = session.query(ReportSQL).filter(ReportSQL.id == id).first()
        name = rp.name
        session.delete(rp)
        return [{"name": "", "type": "p", "value": "Success!", "required": False, "readonly": False, "hidden": False},
                {"name": "", "type": "p", "value": "Successfully deleted report: " + name, "required": False, "readonly": False, "hidden": False}]



from six import with_metaclass
from abc import abstractmethod, ABCMeta


class BaseStorage(with_metaclass(ABCMeta)):
    '''Abstract base class representing the interface to a storage configuration'''
    def __init__(self, config, *args, **kwargs):
        '''
        Args:
            config: Paperboy application config
        '''
        self.config = config

    @abstractmethod
    def status(self, *args, **kwargs):
        '''Method to populate general status information in Status tab of gui'''
        pass

    @abstractmethod
    def form(self, *args, **kwargs):
        '''Method to generate form on gui to create new instance.

        Returns:
            paperboy.config.forms.Response as dict
        '''
        pass

    @abstractmethod
    def search(self, *args, **kwargs):
        '''Method to search for instance'''
        pass

    @abstractmethod
    def list(self, *args, **kwargs):
        '''Method to list all instances of type'''
        pass

    @abstractmethod
    def detail(self, *args, **kwargs):
        '''Method to view detailed fields about type'''
        pass

    @abstractmethod
    def store(self, *args, **kwargs):
        '''Method to save/update instance'''
        pass

    @abstractmethod
    def delete(self, *args, **kwargs):
        '''Method to delete instance'''
        pass


class UserStorage(BaseStorage):
    def __init__(self, config, *args, **kwargs):
        self.config = config

    @abstractmethod
    def login(self, *args, **kwargs):
        '''Handler for user login'''
        pass

    def logout(self, *args, **kwargs):
        '''Handler for user logout'''
        return True


class NotebookStorage(BaseStorage):
    '''Base class for Notebook storage. Notebook backends should inherit from this class'''
    pass


class JobStorage(BaseStorage):
    '''Base class for Job storage. Job backends should inherit from this class'''
    pass


class ReportStorage(BaseStorage):
    '''Base class for Report storage. Report backends should inherit from this class'''
    @abstractmethod
    def generate(self, *args, **kwargs):
        '''Generate reports from Job instance and JSONL papermill parameters'''
        pass



import os
import os.path
from six.moves.urllib_parse import urljoin
from datetime import datetime
from traitlets import HasTraits, Unicode, Instance, Bool
from .forms import Response, FormEntry, DOMEntry
from .base import Base, _REPORT_TYPES, _OUTPUT_TYPES
from .notebook import NotebookConfig
from .job import JobConfig


TEMPLATE_BASEPATH = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'worker', 'nbconvert_templates'))


def _type_to_template(output, strip_code):
    '''select correct nbconvert template for type'''
    if output in ('pdf', 'html'):
        ret = output
        if output == 'pdf':
            ret += '.tplx'
        else:
            ret += '.tpl'
        return ret
    elif output == 'email':
        ret = 'html_email'
        ret += '.tpl'
        return ret
    else:
        return ''


class ReportMetadataConfig(HasTraits):
    '''Paperboy configuration object representing a Report (metadata component)'''
    notebook = Instance(NotebookConfig)
    job = Instance(JobConfig)

    username = Unicode()
    userid = Unicode()

    run = Instance(datetime)
    parameters = Unicode()
    type = Unicode()
    output = Unicode()
    strip_code = Bool()

    template = Unicode()

    run = Instance(datetime, allow_none=True)
    created = Instance(datetime)
    modified = Instance(datetime)

    def to_json(self, include_notebook=False):
        '''Convert ReportMetadata to a JSON'''
        ret = {}
        ret = {}
        ret['notebook'] = self.notebook.name
        if include_notebook:
            ret['notebook_text'] = self.notebook.meta.notebook
        # ret['notebookid'] = self.notebook.id

        ret['job'] = self.job.name
        # ret['jobid'] = self.job.id

        ret['parameters'] = self.parameters
        ret['type'] = self.type
        ret['output'] = self.output
        ret['strip_code'] = self.strip_code

        ret['template'] = self.template or os.path.join(TEMPLATE_BASEPATH, _type_to_template(self.output, self.strip_code))

        if self.run:
            ret['run'] = self.run.strftime('%m/%d/%Y %H:%M:%S')
        else:
            ret['run'] = 'not run'
        ret['created'] = self.created.strftime('%m/%d/%Y %H:%M:%S')
        ret['modified'] = self.modified.strftime('%m/%d/%Y %H:%M:%S')

        return ret

    @staticmethod
    def from_json(jsn):
        '''Create ReportMetadata from a JSON'''
        ret = ReportMetadataConfig()
        for k, v in jsn.items():
            if k in ('created', 'modified', 'run'):
                ret.set_trait(k, datetime.strptime(v, '%m/%d/%Y %H:%M:%S'))
            else:
                ret.set_trait(k, v)
        return ret


class ReportConfig(Base):
    '''Paperboy configuration object representing a Report'''
    name = Unicode()
    id = Unicode()
    meta = Instance(ReportMetadataConfig)

    def to_json(self, include_notebook=False):
        '''Convert Report to a JSON'''
        ret = {}
        ret['name'] = self.name
        ret['id'] = self.id
        ret['meta'] = self.meta.to_json(include_notebook)
        return ret

    def form(self):
        '''Generate Form template for client from a Report object'''
        f = Response()
        f.entries = [
            FormEntry(name='name', type='text', label='Name', placeholder='Name for Report...', required=True),
            FormEntry(name='notebook', type='autocomplete', label='Notebook', url=urljoin(self.config.apiurl, 'autocomplete?type=notebooks&partial='), required=True),
            FormEntry(name='job', type='autocomplete', label='Job', url=urljoin(self.config.apiurl, 'autocomplete?type=jobs&partial='), required=True),
            FormEntry(name='parameters', type='textarea', label='Parameters', placeholder='JSON Parameters...'),
            FormEntry(name='type', type='select', label='Type', options=_REPORT_TYPES, required=True),
            FormEntry(name='output', type='select', label='Output', options=_OUTPUT_TYPES, required=True),
            FormEntry(name='code', type='select', label='Strip Code', options=['yes', 'no'], required=True),
            FormEntry(name='template', type='text', label='Template', required=False),
            FormEntry(name='submit', type='submit', value='save', url=urljoin(self.config.apiurl, 'reports?action=save'))
        ]
        return f.to_json()

    @staticmethod
    def from_json(jsn, config):
        '''Create Report from a JSON'''
        ret = ReportConfig(config)
        ret.name = jsn['name']
        ret.id = jsn['id']
        ret.meta = ReportMetadataConfig.from_json(jsn['meta'])
        ret.meta.notebook = NotebookConfig(config)
        ret.meta.job = JobConfig(config)
        return ret

    def edit(self):
        '''Generate Edit template for client from a Report object'''
        f = Response()
        f.entries = [
            FormEntry(name='name', type='text', value=self.name, label='Name', placeholder='Name for Report...', required=True),
            FormEntry(name='id', type='text', value=self.id, label='Id', hidden=True),
            FormEntry(name='notebook name', type='text', value=self.meta.notebook.name, label='Notebook', required=True, readonly=True),
            FormEntry(name='notebook', type='text', value=self.meta.notebook.id, required=True, readonly=True, hidden=True),
            FormEntry(name='job name', type='text', value=self.meta.job.name, label='Job', required=True, readonly=True),
            FormEntry(name='job', type='text', value=self.meta.job.id, required=True, readonly=True, hidden=True),
            FormEntry(name='parameters', type='textarea', value=self.meta.parameters, label='Parameters', placeholder='JSON Parameters...'),
            FormEntry(name='type', type='select', value=self.meta.type, label='Type', options=_REPORT_TYPES, required=True),
            FormEntry(name='output', type='select', value=self.meta.output, label='Output', options=_OUTPUT_TYPES, required=True),
            FormEntry(name='code', type='select', value='yes' if self.meta.strip_code else 'no', label='Strip Code', options=['yes', 'no'], required=True),
            FormEntry(name='save', type='submit', value='save', url=urljoin(self.config.apiurl, 'reports?action=save')),
            FormEntry(name='delete', type='submit', value='delete', url=urljoin(self.config.apiurl, 'reports?action=delete'))
        ]
        return f.to_json()

    def entry(self):
        '''Generate ListTable entry for client from a Report object'''
        f = Response()
        f.entries = [
            DOMEntry(name='name', type='label', value=self.name, label='Name'),
            DOMEntry(name='id', type='label', value=self.id, label='Id', hidden=True),
            DOMEntry(name='notebook', type='label', value=self.meta.notebook.name, label='Notebook'),
            DOMEntry(name='job', type='label', value=self.meta.job.name, label='Job'),
            DOMEntry(name='parameters', type='json', value=self.meta.parameters, label='Parameters'),
            DOMEntry(name='type', type='label', value=self.meta.type, label='Type'),
            DOMEntry(name='output', type='label', value=self.meta.output, label='Output'),
            DOMEntry(name='strip_code', type='label', value=str(self.meta.strip_code), label='Strip Code'),
            DOMEntry(name='template', type='label', value=self.meta.template, label='Template'),
            # DOMEntry(type='label', value=self.meta.run, label='Run'),
            DOMEntry(name='created', type='label', value=self.meta.created.strftime('%m/%d/%Y %H:%M:%S'), label='Created'),
            DOMEntry(name='modified', type='label', value=self.meta.modified.strftime('%m/%d/%Y %H:%M:%S'), label='Modified'),
            DOMEntry(name='delete', type='button', value='delete', label='delete', url=urljoin(self.config.apiurl, 'reports?action=delete'))
        ]
        return f.to_json()

    def store(self):
        '''Generate response modal for client when saving a Report object'''
        ret = Response()
        ret.entries = [
            DOMEntry(type='h2', value='Success!'),
            DOMEntry(type='p', value='Successfully configured report: {}!'.format(self.name)),
            DOMEntry(type='p', value='Notebook: {}'.format(self.meta.notebook.name)),
            DOMEntry(type='p', value='Job: {}'.format(self.meta.job.name)),
        ]
        return ret.to_json()




class NoUserMiddleware(object):
    '''Dummy user authentication middleware'''
    def __init__(self, config, db, *args, **kwargs):
        self.config = config
        self.db = db

    def process_request(self, req, resp):
        '''inject anonymous user into every context'''
        from paperboy.config import User
        req.context['user'] = User(self.config, id='1', name='anon')


class NoAuthRequiredMiddleware(object):
    '''Dummy no-auth-required middleware'''
    def __init__(self, config, db, *args, **kwargs):
        self.config = config
        self.db = db

    def process_resource(self, req, resp, resource, params):
        pass



from sqlalchemy import Column, Integer, String, DateTime
from sqlalchemy import ForeignKey
from sqlalchemy.orm import relationship
from paperboy.config import NotebookConfig, NotebookMetadataConfig
from .base import Base


class NotebookSQL(Base):
    __tablename__ = 'notebooks'
    id = Column(Integer, primary_key=True)
    name = Column(String)

    userId = Column(Integer, ForeignKey('users.id'))
    user = relationship('UserSQL', back_populates='notebooks')

    jobs = relationship('JobSQL', cascade='all,delete,delete-orphan', back_populates='notebook')
    reports = relationship('ReportSQL', cascade='all,delete,delete-orphan', back_populates='notebook')

    notebook = Column(String)
    privacy = Column(String, nullable=True)
    level = Column(String, nullable=True)
    requirements = Column(String, nullable=True)
    dockerfile = Column(String, nullable=True)

    created = Column(DateTime)
    modified = Column(DateTime)

    @staticmethod
    def from_config(nb):
        # FIXME
        return NotebookSQL(name=nb.name,
                           userId=int(nb.user.id),
                           notebook=nb.notebook,
                           privacy=nb.meta.privacy,
                           level=nb.meta.level,
                           requirements=nb.meta.requirements,
                           dockerfile=nb.meta.dockerfile,
                           created=nb.meta.created,
                           modified=nb.meta.modified)

    def to_config(self, config):
        ret = NotebookConfig(config)
        ret.id = 'Notebook-' + str(self.id)
        ret.name = self.name

        meta = NotebookMetadataConfig()

        meta.username = self.user.name
        meta.userid = 'User-' + str(self.user.id)

        meta.notebook = self.notebook
        meta.privacy = self.privacy
        meta.level = self.level

        meta.requirements = self.requirements
        meta.dockerfile = self.dockerfile

        meta.jobs = len(self.jobs)
        meta.reports = len(self.reports)

        meta.created = self.created
        meta.modified = self.modified

        ret.meta = meta
        return ret

    def __repr__(self):
        return "<Notebook(name='{}', user='{}', privacy='{}', level='{}'>".format(self.name, self.user, self.privacy, self.level)



import falcon
import jinja2
from .base import BaseResource
from .html import read


class LogoutResource(BaseResource):
    '''Falcon resource for user authentication'''

    auth_required = False

    def __init__(self, *args, **kwargs):
        super(LogoutResource, self).__init__(*args, **kwargs)

    def on_get(self, req, resp):
        '''Get logout page'''
        resp.content_type = 'text/html'
        file = read('logout.html')
        tpl = jinja2.Template(file).render(baseurl=self.config.baseurl,
                                           apiurl=self.config.apiurl,
                                           loginurl=self.config.loginurl,
                                           logouturl=self.config.logouturl)
        resp.body = tpl

    def on_post(self, req, resp):
        '''Log user out using authentication backend'''
        ret = self.db.users.logout(req.context.get('user'), req.params, self.session)

        if req.context.get('user'):
            # remove user from context
            req.context['user'] = None

        if req.context.get('auth_token'):
            # remove auth token from context
            req.context['auth_token'] = None

        if ret:
            # delete cookie and refresh page to base
            resp.unset_cookie('auth_token')
            resp.status = falcon.HTTP_302
            resp.set_header('Location', self.config.baseurl)



from six.moves.urllib_parse import urljoin
from traitlets import Unicode
from .forms import Response, FormEntry, DOMEntry
from .base import Base


class UserConfig(Base):
    '''Paperboy configuration object representing a User'''
    name = Unicode()
    id = Unicode()

    def to_json(self):
        '''Convert User to a JSON'''
        ret = {}
        ret['name'] = self.name
        ret['id'] = self.id
        return ret

    def form(self):
        '''Generate Form template for client from a User object'''
        f = Response()
        f.entries = [
            FormEntry(name='name', type='text', label='Name', placeholder='Name for Notebook...', required=True),
            FormEntry(name='submit', type='submit', value='save', url=urljoin(self.config.apiurl, 'users?action=save')),
        ]
        return f.to_json()

    @staticmethod
    def from_json(jsn, config):
        '''Create User from a JSON'''
        ret = UserConfig(config)
        ret.name = jsn.pop('name')
        ret.id = jsn.pop('id')
        return ret

    def edit(self):
        '''Generate Edit template for client from a User object'''
        f = Response()
        f.entries = [
            FormEntry(name='name', type='text', value=self.name, placeholder='Name for Job...', required=True),
            FormEntry(name='save', type='submit', value='save', url=urljoin(self.config.apiurl, 'users?action=save'))
        ]
        return f.to_json()

    def store(self):
        '''Generate response modal for client when saving a User object'''
        ret = Response()
        ret.entries = [
            DOMEntry(type='p', value='Success!'),
            DOMEntry(type='p', value='Successfully stored user {}'.format(self.name)),
        ]
        return ret



# This is the jinja2 templatized DAG that generates
# the job operator. It has the following form as an
# example:
#
#    Job -> Report -> Papermill -> NBConvert ->      ReportPost -\
#      \--> Report -> Papermill -> Voila -> Dokku -> ReportPost --\
#       \-> Report -> Papermill -> NBConvert ->      ReportPost ----> Job Cleanup
#

import json
from base64 import b64decode

# Base Job Operators
from paperboy.scheduler.airflow_operators import JobOperator, JobCleanupOperator

# Base Report Operators
from paperboy.scheduler.airflow_operators import ReportOperator, ReportPostOperator

# Convert Operators
from paperboy.scheduler.airflow_operators import PapermillOperator, NBConvertOperator

# Publish operators
from paperboy.scheduler.airflow_operators import VoilaOperator, DokkuOperator

from airflow import DAG
from datetime import timedelta, datetime


###################################
# Default arguments for operators #
###################################

# DAG args: https://airflow.incubator.apache.org/code.html?highlight=dag#airflow.models.DAG
dag_args = {
    'description': u'',
    'schedule_interval': '{{interval}}',
    'start_date': datetime.strptime('{{start_date}}', '%m/%d/%Y %H:%M:%S'),
    'end_date': None,
    'full_filepath': None,
    'template_searchpath': None,
    'user_defined_macros': None,
    'user_defined_filters': None,
    'concurrency': 16,
    'max_active_runs': 16,
    'dagrun_timeout': None,
    'sla_miss_callback': None,
    'default_view': u'graph',
    'orientation': 'LR',
    'catchup': False,
    'on_success_callback': None,
    'on_failure_callback': None,
    # 'params': None,
}

# Operator args: https://airflow.incubator.apache.org/code.html#baseoperator
default_operator_args = {
    'owner': '{{owner}}',
    'email': ['{{email}}'],
    'email_on_retry': False,
    'email_on_failure': True,
    'retries': 0,
    'retry_delay': timedelta(minutes=1),
    'retry_exponential_backoff': False,
    'max_retry_delay': None,
    'start_date': None,
    'end_date': None,
    'schedule_interval': None,
    'depends_on_past': False,
    'wait_for_downstream': False,
    # 'params': None,
    'default_args': None,
    'adhoc': False,
    'priority_weight': 1,
    'weight_rule': u'downstream',
    'queue': 'default',
    'pool': None,
    'sla': None,
    'execution_timeout': None,
    'on_failure_callback': None,
    'on_success_callback': None,
    'on_retry_callback': None,
    'trigger_rule': u'all_success',
    'resources': None,
    'run_as_user': None,
    'task_concurrency': None,
    'executor_config': None,
    'inlets': None,
    'outlets': None,
}

###################################
# Inline job and reports as b64 json  #
###################################
job_json = json.loads(b64decode({{job_json}}))
reports_json = json.loads(b64decode({{report_json}}))

###################################
# Create dag from job and reports #
###################################

# The top level dag, representing a Job run on a Notebook
dag = DAG('DAG-' + str(job_json['id']), default_args=default_operator_args, **dag_args)

# The Job operator, used for bundling groups of reports,
# setting up env/image
job = JobOperator(job=job_json, task_id='Job-{}'.format(job_json['id']), dag=dag)

# The cleanup operator, run after all reports are finished
cleanup = JobCleanupOperator(job=job_json, task_id='JobCleanup-{}'.format(job_json['id']), dag=dag)

for rep in reports_json:
    # copy over notebook text (only store 1 copy in the job json)
    rep['meta']['notebook_text'] = job_json['meta']['notebook_text']

    # type is "convert" for nbconversion jobs and "publish" for publish jobs
    type = rep['meta']['type']

    # Report operator, performs the required
    # steps prior to running the report
    r = ReportOperator(report=rep,
                       task_id='Report-{}'.format(rep['id']),
                       dag=dag)
    job.set_downstream(r)

    # Papermill operator, performs the report creation
    # using papermill and the report's individual
    # parameters and configuration
    pp = PapermillOperator(report=rep,
                           task_id='ReportPapermill-{}'.format(rep['id']),
                           dag=dag)
    r.set_downstream(pp)

    # The post-report operator, used for post-report
    # tasks such as sending the report in an email
    # or pushing the deploy to dokku
    rp = ReportPostOperator(report=rep,
                            config=json.loads('{{output_config}}'),
                            task_id='ReportPost-{}'.format(rep['id']), dag=dag)
    rp.set_downstream(cleanup)

    if type == 'convert':
        # NBConvert operator, performs the NBConversion if
        # required
        nb = NBConvertOperator(report=rep,
                               task_id='ReportNBConvert-{}'.format(rep['id']),
                               dag=dag)
        pp.set_downstream(nb)
        nb.set_downstream(rp)

    elif type == 'publish':
        # Assemble a Voila Job from papermilled notebook
        v = VoilaOperator(report=rep,
                          task_id='ReportVoila-{}'.format(rep['id']),
                          dag=dag)
        pp.set_downstream(v)

        # Dokku deployment operator, performs the
        # deploy to the dokku repo
        d = DokkuOperator(report=rep,
                          task_id='ReportDokku-{}'.format(rep['id']),
                          dag=dag)
        v.set_downstream(d)
        d.set_downstream(rp)

    else:
        raise NotImplementedError()



import nbformat
import logging
from datetime import datetime
from paperboy.config import NotebookConfig
from paperboy.storage import NotebookStorage
from sqlalchemy import or_
from .base import BaseSQLStorageMixin, justid
from .models.user import UserSQL
from .models.notebook import NotebookSQL
from ..utils import strip_outputs


class NotebookSQLStorage(BaseSQLStorageMixin, NotebookStorage):
    def status(self, user, params, session, *args, **kwargs):
        base = session.query(NotebookSQL) \
            .filter(or_(NotebookSQL.userId.like((user.id)),
                        (hasattr(NotebookSQL, 'privacy') and NotebookSQL.privacy == 'public')))

        return {'total': base.count(),
                'public': base.filter(NotebookSQL.level == 'public').count(),
                'private': base.filter(NotebookSQL.privacy == 'private').count()}

    def form(self):
        '''Pass through to shared method in BaseSQLStorageMixin'''
        return self._form(NotebookConfig)

    def search(self, user, params, session, *args, **kwargs):
        '''Pass through to shared method in BaseSQLStorageMixin'''
        return self._search(NotebookSQL, 'Notebook', user, params, session, *args, **kwargs)

    def list(self, user, params, session, *args, **kwargs):
        '''Pass through to shared method in BaseSQLStorageMixin'''
        return self._list(NotebookSQL, 'notebooks', user, params, session, *args, **kwargs)

    def detail(self, user, params, session, *args, **kwargs):
        '''Pass through to shared method in BaseSQLStorageMixin'''
        return self._detail(NotebookSQL, user, params, session, *args, **kwargs)

    def store(self, user, params, session, *args, **kwargs):
        name = params.get('name')
        user_sql = session.query(UserSQL).get(int(user.id))

        if params.get('file') is not None:
            notebook = nbformat.writes(strip_outputs(nbformat.reads(params.get('file').file.read(), 4)))
        elif params.get('notebook'):
            notebook = nbformat.writes(strip_outputs(nbformat.reads(params.get('notebook'), 4)))
        else:
            raise NotImplementedError()

        privacy = params.get('privacy') or ''
        level = params.get('level') or ''

        if params.get('requirements') is not None:
            try:
                requirements = params.get('requirements').file.read()
            except AttributeError:
                requirements = ''
        else:
            requirements = ''

        if params.get('dockerfile'):
            try:
                params.get('dockerfile').file.read()
            except AttributeError:
                requirements = ''

        else:
            dockerfile = ''

        created = datetime.now()
        modified = datetime.now()

        id = params.get('id')
        if id:
            id = justid(id)
            nb = session.query(NotebookSQL).filter(NotebookSQL.id == id).first()
            nb.name = name
            nb.notebook = notebook
            nb.privacy = privacy
            nb.level = level
            nb.requirements = requirements
            nb.dockerfile = dockerfile
            nb.modified = modified

        else:
            nb = NotebookSQL(name=name,
                             userId=int(user.id),
                             user=user_sql,
                             notebook=notebook,
                             privacy=privacy,
                             level=level,
                             requirements=requirements,
                             dockerfile=dockerfile,
                             created=created,
                             modified=modified)

        session.add(nb)

        # generate id
        session.flush()
        session.refresh(nb)

        store = nb.to_config(self.config).store()
        logging.critical("Storing notebook {}".format(nb))
        return store

    def delete(self, user, params, session, *args, **kwargs):
        # TODO only if allowed
        id = justid(params.get('id'))
        nb = session.query(NotebookSQL).filter(NotebookSQL.id == id).first()
        name = nb.name
        session.delete(nb)

        # cascade handled in DB, need to cascade to airflow jobs
        from .models.job import JobSQL
        jobs = session.query(JobSQL).filter(JobSQL.notebookId == nb.id).all()
        for job in jobs:
            resp = self.db.jobs.delete(user, {'id': job.id}, session, *args, **kwargs)
            print(resp)
        return [{"name": "", "type": "p", "value": "Success!", "required": False, "readonly": False, "hidden": False},
                {"name": "", "type": "p", "value": "Successfully deleted notebook: " + name, "required": False, "readonly": False, "hidden": False}]



from .common import BaseTask
import json
from paperboy.utils import name_to_class
from luigi.parameter import Parameter, ParameterVisibility


class DokkuTask(BaseTask):
    report = Parameter(visibility=ParameterVisibility.HIDDEN)
    config = Parameter()

    def __init__(self, *args, **kwargs):
        super(DokkuTask, self).__init__(*args, **kwargs)
        config = json.loads(kwargs.get('config', {}))
        self._config = name_to_class(config.get('config')).from_json(config)
        self._report = json.loads(self.report)

    def run(self):
        self.log.critical('report-post')

        fp = self.input().open('r')
        output_nb = fp.read()
        fp.close()

        self.log.critical(output_nb)
        outputter = self.config.clazz(self._config)
        outputter.write(self._report, output_nb, task_id=self.task_id)



###############################
# Remote Airflow              #
# This is used when running   #
# airflow on a remote machine #
###############################
import falcon
import json
import logging
import os
import subprocess
from paperboy.server.deploy import FalconDeploy
from paperboy.config.user import UserConfig
from paperboy.config.notebook import NotebookConfig
from paperboy.config.job import JobConfig
from paperboy.config.report import ReportConfig
from paperboy.config.scheduler import AirflowSchedulerConfig
from paperboy.scheduler.airflow import AirflowScheduler
from six.moves.urllib_parse import urljoin
from sqlalchemy.exc import OperationalError
from traitlets.config.application import Application
from traitlets import Int, Unicode


class RemoteAirflow(Application):
    '''Helper to proxy methods through to a remote airflow application.'''
    name = 'remoteairflow'
    description = 'remoteairflow'

    ############
    # Gunicorn #
    ############
    workers = Int(default_value=1, help="Number of gunicorn workers").tag(config=True)
    port = Unicode(default_value='8080', help="Port to run on").tag(config=True)
    ############

    #############
    # Scheduler #
    #############
    # FIXME doesnt allow default_value yet
    scheduler = AirflowSchedulerConfig()
    #############

    def start(self):
        """Start the whole thing"""
        self.port = os.environ.get('PORT', self.port)
        options = {
            'bind': '0.0.0.0:{}'.format(self.port),
            'workers': self.workers
        }

        def from_base(url):
            return urljoin(self.baseurl, url)

        api = falcon.API()

        remote = RemoteAirflowResource()
        status = RemoteAirflowStatusResource()
        api.add_route(from_base('remote'), remote)
        api.add_route(from_base('status'), status)

        ##########
        port = 8081
        options = {
            'bind': '0.0.0.0:{}'.format(port),
            'workers': 1
        }
        logging.debug('Running on port:{}'.format(port))
        FalconDeploy(api, options).run()

    @classmethod
    def launch_instance(cls, argv=None, **kwargs):
        """Launch an instance of a Paperboy Application"""
        return super(RemoteAirflow, cls).launch_instance(argv=argv, **kwargs)

    aliases = {
        'workers': 'RemoteAirflow.workers',
        'port': 'RemoteAirflow.port',
        'baseurl': 'RemoteAirflow.baseurl',
    }


class RemoteAirflowResource(object):
    def __init__(self):
        pass

    def on_get(self, req, resp):
        resp.content_type = 'application/json'
        resp.body = json.dumps({'status': 'ok'})

    def on_post(self, req, resp):
        name = req.params['name']
        user = UserConfig.from_config(req.params['user'], self.config)
        notebook = NotebookConfig.from_config(req.params['notebook'], self.config)
        job = JobConfig.from_config(req.params['job'], self.config)
        reports = [ReportConfig.from_config(r, self.config) for r in req.params['reports']]

        if reports:
            # write or rewrite
            template = AirflowScheduler.template(self.config, user, notebook, job, reports)
            name = job.id + '.py'
            with open(os.path.join(self.config.scheduler.dagbag, name), 'w') as fp:
                fp.write(template)

            resp.content_type = 'application/json'
            resp.body = json.dumps({'status': 'ok'})
        else:
            # delete
            name = job.id + '.py'
            file = os.path.join(self.config.scheduler.dagbag, name)
            dag = 'DAG-' + job.id

            # delete dag file
            os.remove(file)

            # delete dag
            # FIXME
            try:
                cmd = ['airflow', 'delete_dag', dag, '-y']
                subprocess.call(cmd)
            except Exception as e:
                logging.error(e)


class RemoteAirflowStatusResource(object):
    def __init__(self):
        pass

    def on_get(self, req, resp):
        # TODO pull status args out of request
        engine = req.params.get('engine')
        type = req.params.get('type', '')
        try:
            gen = AirflowScheduler.query(engine)
        except OperationalError:
            logging.debug('Scheduler offline, using fake scheduler query')
            gen = AirflowScheduler.fakequery(engine)
        if type == 'jobs':
            ret = gen['jobs']
        elif type == 'reports':
            ret = gen['reports']
        else:
            ret = gen

        resp.content_type = 'application/json'
        resp.body = json.dumps(ret)

    def on_post(self, req, resp):
        resp.content_type = 'application/json'
        resp.body = json.dumps({'status': 'ok'})


if __name__ == '__main__':
    RemoteAirflow.launch_instance()



def class_to_name(cls):
    return '.'.join((cls.__module__, cls.__qualname__))


def name_to_class(name):
    components = name.split('.')
    mod = __import__(components[0])
    for comp in components[1:]:
        mod = getattr(mod, comp)
    return mod



import falcon
from six.moves.urllib_parse import urljoin
from ..resources import StaticResource, HTMLResource, LoginResource, LogoutResource, RegisterResource
from ..resources import StatusResource, AutocompleteResource, ConfigResource, SchedulerResource
from ..resources import NotebookResource, JobResource, ReportResource
from ..resources import NotebookDetailResource, JobDetailResource, ReportDetailResource
from ..storage import StorageEngine, StorageError


def FalconAPI(config):
    '''Create falcon.API application from Paperboy traitlets application

    Args:
        config (paperboy.config.application.Paperboy): traitlets configuration for application

    Returns:
        falcon.API: the routed falcon api to launch with gunicorn/waitress
    '''
    def from_base(url):
        return urljoin(config.baseurl, url)

    def from_api(url):
        return urljoin(config.apiurl, url)

    ###########
    # Storage #
    ###########
    db = StorageEngine(config.storage.user_storage(config), config.storage.notebook_storage(config), config.storage.job_storage(config), config.storage.report_storage(config))
    api = falcon.API(middleware=[config.auth_required_middleware(config, db),
                                 config.load_user_middleware(config, db)] +
                     config.essential_middleware +
                     config.extra_middleware)
    api.add_error_handler(StorageError, StorageError.handle)

    #############
    # Scheduler #
    #############
    scheduler = config.scheduler_config.clazz(config, db)

    ####################
    # Static resources #
    ####################
    html = HTMLResource(config)
    static = StaticResource(config)
    api.add_route(from_base(''), html)
    api.add_route(from_base('index.html'), html)
    api.add_sink(static.on_get, prefix=from_base('static'))

    ##################
    # Auth Resources #
    ##################
    login = LoginResource(config, db)
    api.add_route(from_base(config.loginurl), login)
    logout = LogoutResource(config, db)
    api.add_route(from_base(config.logouturl), logout)
    register = RegisterResource(config, db)
    api.add_route(from_base(config.registerurl), register)

    ##########
    # Routes #
    ##########
    kwargs = {'config': config,
              'db': db,
              'scheduler': scheduler}

    # Status
    status = StatusResource(**kwargs)
    api.add_route(from_api('status'), status)

    # Scheduler
    schedulerresource = SchedulerResource(**kwargs)
    api.add_route(from_api('scheduler'), schedulerresource)

    # Autocomplete
    autocomplete = AutocompleteResource(**kwargs)
    api.add_route(from_api('autocomplete'), autocomplete)

    # Config
    configresource = ConfigResource(**kwargs)
    api.add_route(from_api('config'), configresource)

    # Notebooks
    notebooks = NotebookResource(**kwargs)
    api.add_route(from_api('notebooks'), notebooks)

    notebooksdetail = NotebookDetailResource(**kwargs)
    api.add_route(from_api('notebooks/details'), notebooksdetail)

    # Jobs
    jobs = JobResource(**kwargs)
    api.add_route(from_api('jobs'), jobs)

    jobdetail = JobDetailResource(**kwargs)
    api.add_route(from_api('jobs/details'), jobdetail)

    # Reports
    reports = ReportResource(**kwargs)
    api.add_route(from_api('reports'), reports)

    reportdetail = ReportDetailResource(**kwargs)
    api.add_route(from_api('reports/details'), reportdetail)

    # Extra handlers
    for route, handler in config.extra_handlers:
        api.add_route(route, handler(**kwargs))
    return api



from waitress import serve


class FalconWaitress():
    '''Utility to deploy falcon.API on waitress'''
    def __init__(self, app, options=None):
        '''Constructor
        Args:
            app (falcon.API): a routed falcon API
            options (dict): a set of options for gunicorn (e.g. workers, port, etc)
        '''
        self.application = app
        self.host, self.port = options.get('bind', '0.0.0.0:8080').split(':')
        self.workers = options.get('workers', 1)
        super(FalconWaitress, self).__init__()

    def run(self):
        serve(self.application, host=self.host, port=self.port, threads=self.workers)



import json
from .base import BaseResource


class ReportResource(BaseResource):
    def __init__(self, *args, **kwargs):
        super(ReportResource, self).__init__(*args, **kwargs)

    def on_get(self, req, resp):
        '''List all report instances'''
        resp.content_type = 'application/json'
        resp.body = json.dumps(self.db.reports.list(req.context['user'], req.params, self.session))

    def on_post(self, req, resp):
        '''Create new or delete old report instance'''
        resp.content_type = 'application/json'
        action = req.params.get('action')
        if action == 'delete':
            resp.body = json.dumps(self.db.reports.delete(req.context['user'], req.params, self.session))
        else:
            resp.body = json.dumps(self.db.reports.store(req.context['user'], req.params, self.session))


class ReportDetailResource(BaseResource):
    def __init__(self, *args, **kwargs):
        super(ReportDetailResource, self).__init__(*args, **kwargs)

    def on_get(self, req, resp):
        '''Get details of specific report instance'''
        resp.content_type = 'application/json'
        resp.body = json.dumps(self.db.reports.detail(req.context['user'], req.params, self.session))



'''Copied from https://gitlab.com/skosh/falcon-helpers due to conda-forge issues
Copyright (c) 2017 by Nicholas Zaccardi

Some rights reserved.

Redistribution and use in source and binary forms of the software as well
as documentation, with or without modification, are permitted provided
that the following conditions are met:

* Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above
  copyright notice, this list of conditions and the following
  disclaimer in the documentation and/or other materials provided
  with the distribution.

* The names of the contributors may not be used to endorse or
  promote products derived from this software without specific
  prior written permission.

THIS SOFTWARE AND DOCUMENTATION IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE AND DOCUMENTATION, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
'''
from sqlalchemy.orm import sessionmaker, scoped_session
_session_maker = sessionmaker()
_session = scoped_session(_session_maker)


class LoadUserMiddleware:
    """Load a user from the database during the request cycle using sqlalchemy

    By default this will grab the id data from the auth token on the request
    context assuming you are using `.middleware.auth_required`. To change that
    pass a function which takes the request and returns the user id to `get_id`.

    :param session: a sqlalchemy session
    :param user_cls: a class which will return the user object. This object must
        have a `get_by_id` which will return a SQLAlchemy Query.
    :param get_id: a function which will get the user identifier off of the
        request.

    """

    def __init__(self, user_cls, get_id=None, session=None):
        self.user_cls = user_cls
        self.get_id = get_id or self._get_id
        self.session = session or _session

    @staticmethod
    def _get_id(req):
        try:
            return req.context.get('auth_token_contents').get('sub')
        except AttributeError:
            return None

    def fetch_user(self, user_id):
        return (self.user_cls
                    .get_by_id(user_id)
                    .with_session(self.session())
                    .one_or_none())

    def process_request(self, req, resp):
        user_id = self.get_id(req)
        req.context['user'] = self.fetch_user(user_id)



import falcon
import logging
import os
from six.moves.urllib_parse import urljoin
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from traitlets.config.application import Application
from traitlets import Int, Instance, List, Tuple, Unicode, Bool, validate, TraitError
from uuid import uuid4

# falcon api
from ..server.api import FalconAPI

# deployer
from ..server.deploy import FalconDeploy

# base configs
from .user import UserConfig
from .notebook import NotebookConfig
from .job import JobConfig
from .report import ReportConfig
from .scheduler import SchedulerConfig, DummySchedulerConfig, AirflowSchedulerConfig, LuigiSchedulerConfig, LocalSchedulerConfig
from .storage import SQLAStorageConfig
from .output import LocalOutputConfig

# no auth
from ..middleware import NoUserMiddleware, NoAuthRequiredMiddleware

# essential middleware
from ..middleware import CORSMiddleware, MultipartMiddleware

# sql
from ..storage.sqla import Base
from ..storage.sqla import UserSQLStorage, NotebookSQLStorage, JobSQLStorage, ReportSQLStorage
from ..middleware import SQLAlchemySessionMiddleware, SQLUserMiddleware, SQLAuthRequiredMiddleware


class Paperboy(Application):
    """Base class for paperboy applications"""
    name = 'paperboy'
    description = 'paperboy'

    ############
    # Gunicorn #
    ############
    workers = Int(default_value=1, help="Number of gunicorn workers").tag(config=True)
    port = Unicode(default_value='8080', help="Port to run on").tag(config=True)
    ############

    ##########
    # Falcon #
    ##########
    api = Instance(falcon.API, help="A Falcon API instance").tag(config=True)
    ##########

    ########
    # URLs #
    ########
    baseurl = Unicode(default_value='/', help="Base URL (for reverse proxies)").tag(config=True)
    apiurl = Unicode(default_value='/api/v1/', help="API base URL (for reverse proxies)").tag(config=True)
    loginurl = Unicode(default_value='login', help="login url").tag(config=True)
    logouturl = Unicode(default_value='logout', help="logout url").tag(config=True)
    registerurl = Unicode(default_value='register', help="register url").tag(config=True)
    ########

    ########
    # Auth #
    ########
    http = Bool(default_value=True, help="Running on HTTP (as opposed to https, so token is insecure)").tag(config=True)
    include_password = Bool(default_value=False).tag(config=True)
    include_register = Bool(default_value=True).tag(config=True)
    token_timeout = Int(default_value=600).tag(config=True)
    #############

    ##########
    # Config #
    ##########
    # FIXME doesnt allow default_value yet
    user_config = UserConfig
    notebook_config = NotebookConfig
    job_config = JobConfig
    report_config = ReportConfig
    ##########

    ##############
    # Middleware #
    ##############
    essential_middleware = [CORSMiddleware(allow_all_origins=True).middleware,
                            MultipartMiddleware()]
    extra_middleware = List(default_value=[])  # List of extra middlewares to install
    auth_required_middleware = Instance(object)
    load_user_middleware = Instance(object)
    ##############

    ##################
    # Custom handler #
    ##################
    extra_handlers = List(trait=Tuple(), default_value=[])  # List of tuples (route, handler) of handlers to install
    ##################

    ##########################################
    #        Predefined Configurations       #
    #
    ##########################################
    backend = Unicode(default_value='dummy', help="Backend set to use, options are {sqla, custom}").tag(config=True)
    scheduler = Unicode(default_value='dummy', help="Scheduler type to use, options are {dummy, airflow, luigi}").tag(config=True)
    auth = Unicode(default_value='dummy', help="Authentication backend set to use, options are {none, sqla, custom}").tag(config=True)
    secret = Unicode()

    @validate('backend')
    def _validate_backend(self, proposed):
        if proposed['value'] not in ('custom', 'dummy', 'git', 'sqla',):
            raise TraitError('backend not recognized: %s'.format(proposed['value']))
        return proposed['value']

    @validate('auth')
    def _validate_auth(self, proposed):
        if proposed['value'] not in ('custom', 'none', 'sqla',):
            raise TraitError('backend not recognized: %s'.format(proposed['value']))
        return proposed['value']
    ##########################################

    ###########
    # Storage #
    ###########
    # FIXME doesnt allow default_value yet
    storage = SQLAStorageConfig()
    dev = Bool(default_value=False)
    ###########

    #############
    # Scheduler #
    #############
    # FIXME doesnt allow default_value yet
    scheduler_config = Instance(klass=SchedulerConfig, args=(), kwargs={})
    #############

    ##################
    # Output         #
    ##################
    output = Instance(klass=LocalOutputConfig, args=(), kwargs={})
    ##################

    def start(self):
        """Start the whole thing"""
        self.port = os.environ.get('PORT', self.port)

        options = {
            'bind': '0.0.0.0:{}'.format(self.port),
            'workers': self.workers
        }
        self.secret = str(uuid4())

        if self.dev:
            self.sql_url = 'sqlite:///:memory:'
            logging.critical('Using SQL in memory backend')

            self.storage.engine = create_engine(self.storage.sql_url, echo=False)
            Base.metadata.create_all(self.storage.engine)

            self.sessionmaker = sessionmaker(bind=self.storage.engine)
            self.backend = 'sqla'
            self.auth = 'sqla'
            self.extra_middleware = self.extra_middleware + [SQLAlchemySessionMiddleware(self.storage.sessionmaker)]
            self.storage.notebook_storage = NotebookSQLStorage
            self.storage.job_storage = JobSQLStorage
            self.storage.report_storage = ReportSQLStorage
            self.storage.user_storage = UserSQLStorage
            self.storage.sql_user = True

            logging.critical('Using SQL auth')
            self.auth_required_middleware = SQLAuthRequiredMiddleware
            self.load_user_middleware = SQLUserMiddleware

            logging.critical('Using Dummy scheduler')
            self.scheduler = 'dummy'
            self.scheduler_config = DummySchedulerConfig()

        else:

            # Preconfigured storage backends
            if self.backend == 'git':
                logging.critical('Using Git backend')
                raise NotImplementedError()

            # default to sqla
            # elif self.backend == 'sqla':
            else:
                logging.critical('Using SQL backend')

                self.storage.engine = create_engine(os.environ.get('PAPERBOY_SQL_URL') or self.storage.sql_url, echo=False)
                Base.metadata.create_all(self.storage.engine)

                self.storage.sessionmaker = sessionmaker(bind=self.storage.engine)
                self.extra_middleware = self.extra_middleware + [SQLAlchemySessionMiddleware(self.storage.sessionmaker)]
                self.storage.notebook_storage = NotebookSQLStorage
                self.storage.job_storage = JobSQLStorage
                self.storage.report_storage = ReportSQLStorage
                self.storage.user_storage = UserSQLStorage
                self.storage.sql_user = True
                self.auth = 'sqla'

            # Preconfigured auth backends
            if self.auth == 'none':
                logging.critical('Using No auth')
                self.auth_required_middleware = NoAuthRequiredMiddleware
                self.load_user_middleware = NoUserMiddleware

            elif self.auth == 'sqla':
                logging.critical('Using SQL auth')
                self.auth_required_middleware = SQLAuthRequiredMiddleware
                self.load_user_middleware = SQLUserMiddleware

            if self.scheduler == 'dummy':
                logging.critical('Using dummy scheduler')
                self.scheduler_config = DummySchedulerConfig()

            elif self.scheduler == 'airflow':
                logging.critical('Using Airflow scheduler')
                self.scheduler_config = AirflowSchedulerConfig()

            elif self.scheduler == 'luigi':
                logging.critical('Using Luigi scheduler')
                self.scheduler_config = LuigiSchedulerConfig()

            elif self.scheduler == 'local':
                logging.critical('Using Local scheduler')
                self.scheduler_config = LocalSchedulerConfig()

            else:
                raise Exception('Must specify a valid scheduler!')

        FalconDeploy(FalconAPI(self), options).run()

    @classmethod
    def launch_instance(cls, argv=None, **kwargs):
        """Launch an instance of a Paperboy Application"""
        return super(Paperboy, cls).launch_instance(argv=argv, **kwargs)

    def to_dict(self):
        return {'name': self.name,
                'description': self.description,
                'workers': self.workers,
                'port': self.port}
    aliases = {
        'workers': 'Paperboy.workers',
        'port': 'Paperboy.port',
        'baseurl': 'Paperboy.baseurl',
        'backend': 'Paperboy.backend',
        'scheduler': 'Paperboy.scheduler',
        'auth': 'Paperboy.auth',
        'sql_url': 'Paperboy.storage.sql_url',
    }

    def _login_redirect(config, *args, **kwargs):
        raise falcon.HTTPFound(urljoin(config.baseurl, config.loginurl))



import os

'''Windows does not support gunicorn, so on windows use Waitress'''
if os.name == 'nt':
    from .deploy_win import FalconWaitress as FalconDeploy  # noqa: F401
else:
    from .deploy_nix import FalconGunicorn as FalconDeploy  # noqa: F401



import jwt
import logging
from paperboy.config import UserConfig
from paperboy.storage import UserStorage
from .base import BaseSQLStorageMixin
from .models.user import UserSQL


class UserSQLStorage(BaseSQLStorageMixin, UserStorage):
    def __init__(self, *args, **kwargs):
        super(UserSQLStorage, self).__init__(*args, **kwargs)

    def status(self, *args, **kwargs):
        '''Not used'''
        return {}

    def form(self):
        '''Pass through to shared method in BaseSQLStorageMixin'''
        return self._form(UserConfig)

    def search(self, user, params, session, *args, **kwargs):
        '''Pass through to shared method in BaseSQLStorageMixin'''
        return self._search(UserSQL, 'User', user, params, session, *args, **kwargs)

    def login(self, user, params, session, *args, **kwargs):
        '''username/password -> user/token'''
        username = params.get('username')
        password = params.get('password') or ''
        user = session.query(UserSQL).filter_by(name=username, password=password).first()

        if user:
            token = jwt.encode({'id': str(user.id), 'name': user.name}, self.config.secret, algorithm='HS256').decode('ascii')
            return token

    def list(self, user, params, session, *args, **kwargs):
        '''Not used'''
        return {}

    def detail(self, user, params, session, *args, **kwargs):
        '''token -> user'''
        try:
            user = jwt.decode(user, self.config.secret, algorithms=['HS256'])
        except (jwt.exceptions.InvalidSignatureError, jwt.exceptions.DecodeError):
            return None
        return UserConfig(self.config, name=user['name'], id=user['id'])

    def store(self, user, params, session, *args, **kwargs):
        username = params.get('username')
        password = params.get('password') or ''
        user = UserSQL(name=username, password=password)

        session.add(user)  # may raise exception

        # generate id
        session.flush()
        session.refresh(user)

        token = jwt.encode({'id': str(user.id), 'name': user.name}, self.config.secret, algorithm='HS256').decode('ascii')
        logging.critical("Storing user {} {} {}".format(username, token, user.id))
        return token

    def delete(self, user, params, session, *args, **kwargs):
        user = session.query(UserSQL).filter(UserSQL.id == user.id).first()
        name = user.name
        session.delete(user)
        return [{"name": "", "type": "p", "value": "Success!", "required": False, "readonly": False, "hidden": False},
                {"name": "", "type": "p", "value": "Successfully deleted user: " + name, "required": False, "readonly": False, "hidden": False}]



# -*- coding: utf-8 -*-
import json
import os
import os.path
import jinja2
import sys
import logging
from base64 import b64encode
from ..cron import schedule_cron, unschedule_cron
from ..base import BaseScheduler, interval_to_cron

with open(os.path.abspath(os.path.join(os.path.dirname(__file__), 'paperboy.luigi.py')), 'r') as fp:
    TEMPLATE = fp.read()

LUIGI_URLS = {
    'PENDING': '/api/task_list?data={"status":"PENDING"}',
    'RUNNING': '/api/task_list?data={"status":"RUNNING"}',
    'FAILED': '/api/task_list?data={"status":"FAILED"}',
    'DONE': '/api/task_list?data={"status":"DONE"}',
    'DISABLED': '/api/task_list?data={"status":"DISABLED"}',
}


class LuigiScheduler(BaseScheduler):
    def __init__(self, *args, **kwargs):
        '''Create a new luigi scheduler, connecting to the luigi instances configuration'''
        super(LuigiScheduler, self).__init__(*args, **kwargs)

    def status(self, user, params, session, *args, **kwargs):
        '''Get status of job/report tasks'''
        type = params.get('type', '')

        # TODO
        logging.debug('Scheduler offline, using fake scheduler query')
        gen = LuigiScheduler.query()
        if type == 'jobs':
            return gen.get('jobs', [])
        elif type == 'reports':
            return gen.get('reports', [])
        else:
            return gen or {}

    @staticmethod
    def query():
        '''Get status of job/report tasks from luigi'''
        return {}

    @staticmethod
    def template(config, user, notebook, job, reports, *args, **kwargs):
        '''jinja templatize luigi task for paperboy (paperboy.luigi.py)'''
        owner = user.name
        start_date = job.meta.start_time.strftime('%m/%d/%Y %H:%M:%S')
        email = 'test@test.com'
        job_json = b64encode(json.dumps(job.to_json(True)).encode('utf-8'))
        report_json = b64encode(json.dumps([r.to_json() for r in reports]).encode('utf-8'))
        interval = interval_to_cron(job.meta.interval, job.meta.start_time)

        tpl = jinja2.Template(TEMPLATE).render(
            owner=owner,
            start_date=start_date,
            interval=interval,
            email=email,
            job_json=job_json,
            report_json=report_json,
            output_config=json.dumps(config.output.to_json())
            )
        return tpl

    def schedule(self, user, notebook, job, reports, *args, **kwargs):
        '''Schedule a task for `job` composed of `reports` to be run on luigi'''
        # create task
        template = LuigiScheduler.template(self.config, user, notebook, job, reports, *args, **kwargs)
        name = job.id + '.py'
        os.makedirs(self.config.scheduler_config.task_folder, exist_ok=True)
        with open(os.path.join(self.config.scheduler_config.task_folder, name), 'w') as fp:
            fp.write(template)

        # create crontab
        schedule_cron(self.luigi_command(os.path.join(self.config.scheduler_config.task_folder, name)), interval_to_cron(job.meta.interval, job.meta.start_time), self.config.scheduler_config.crontab)
        return template

    def luigi_command(self, path):
        return ' '.join((sys.executable, path))

    def unschedule(self, user, notebook, job, reports, *args, **kwargs):
        '''Remove the task for `user` and `notebook` composed of `job` running `reports` from
            luigi 2 parts, remove the task from disk and unschedule the task in cron'''
        if reports:
            # reschedule
            return self.schedule(user, notebook, job, reports, *args, **kwargs)

        else:
            # delete
            return unschedule_cron(self.luigi_command(os.path.join(self.config.scheduler_config.task_folder, job.id + '.py')), self.config.scheduler_config.crontab)



import os
import os.path
from datetime import datetime
from .base import BaseOutput


class LocalOutput(BaseOutput):
    '''Output to local filesystem'''
    def __init__(self, config, *args, **kwargs):
        self.config = config

    def write(self, report, output, *args, **kwargs):
        task_id = kwargs.get('task_id', '')
        path = os.path.join(self.config.dir, task_id) + '_' + datetime.now().strftime('%m-%d-%Y_%H-%M-%S')

        if isinstance(report, dict):
            if report['meta']['output'] == 'notebook':
                path += '.ipynb'
            elif report['meta']['output'] == 'script':
                path += '.py'
            elif report['meta']['output'] == 'email':
                path += '.eml'
            elif report['meta']['output'] in ('pdf', 'html'):
                path += '.{}'.format(report['meta']['output'])
        else:
            if report.meta.output == 'notebook':
                path += '.ipynb'
            elif report.meta.output == 'script':
                path += '.py'
            elif report.meta.output == 'email':
                path += '.eml'
            elif report.meta.output in ('pdf', 'html'):
                path += '.{}'.format(report.meta.output)

        if isinstance(output, bytes):
            output = output.decode('utf8')

        with open(path, 'w') as fp:
            fp.write(output)



from enum import Enum
from functools import lru_cache
from traitlets import HasTraits


class BaseEnum(Enum):
    @classmethod
    @lru_cache(None)
    def members(cls):
        return cls.__members__.keys()

    @classmethod
    @lru_cache(None)
    def values(cls):
        return tuple(x.value for x in cls.__members__.values())


class Interval(BaseEnum):
    MINUTELY = 'minutely'
    FIVE_MINUTES = '5 minutes'
    TEN_MINUTES = '10 minutes'
    THIRTY_MINUTES = '30 minutes'
    HOURLY = 'hourly'
    TWO_HOURS = '2 hours'
    THREE_HOURS = '3 hours'
    SIX_HOURS = '6 hours'
    TWELVE_HOURS = '12 hours'
    DAILY = 'daily'
    WEEKLY = 'weekly'
    MONTHLY = 'monthly'


class ReportType(BaseEnum):
    CONVERT = 'convert'
    # PUBLISH = 'publish'


class OutputType(BaseEnum):
    NOTEBOOK = 'notebook'
    PDF = 'pdf'
    HTML = 'html'
    EMAIL = 'email'
    SCRIPT = 'script'
    # POWERPOINT = 'powerpoint'


class PrivacyLevel(BaseEnum):
    PUBLIC = 'public'
    PRIVATE = 'private'


class ServiceLevel(BaseEnum):
    PRODUCTION = 'production'
    RESEARCH = 'research'
    DEVELOPMENT = 'development'
    PERSONAL = 'personal'


_INTERVAL_TYPES = Interval.values()
_REPORT_TYPES = ReportType.values()
_OUTPUT_TYPES = OutputType.values()
_PRIVACY_LEVELS = PrivacyLevel.values()
_SERVICE_LEVELS = ServiceLevel.values()


class Base(HasTraits):
    '''Base HasTraits abstract base class for all paperboy configureables (User, Notebook, Job, and Report)'''

    def __init__(self, config, *args, **kwargs):
        super(Base, self).__init__(*args, **kwargs)
        self.config = config

    @staticmethod
    def from_json(jsn, config):
        '''create a paperboy config object from a json

        Args:
            jsn: python dictionary from json representing the configuration object
            config: paperboy configuration to populate from json

        Returns:
            subclass of Base populated from jsn
        '''
        raise NotImplementedError()

    def to_json(self, include_notebook=False):
        '''convert a paperboy config to a json

        Args:
            self: subclass of Base
            include_notebook: if config would include a notebook (potentially several MB in size), should
                we include it or strip it?

        Returns:
            dict: python dictionary representing the response json
        '''
        raise NotImplementedError()

    def form(self):
        '''generate a JSON form template for the subclass of Base

        Args:
            self: subclass of Base

        Returns:
            dict: python dictionary representing the form template as a JSON
        '''
        raise NotImplementedError()

    def edit(self):
        '''generate a JSON edit template for the subclass of Base

        Args:
            self: subclass of Base

        Returns:
            dict: python dictionary representing the edit template as a JSON
        '''
        raise NotImplementedError()

    def entry(self):
        raise NotImplementedError()

    def store(self):
        raise NotImplementedError()



from .base import *  # noqa: F401, F403
from .forms import *  # noqa: F401, F403
from .job import *  # noqa: F401, F403
from .notebook import *  # noqa: F401, F403
from .output import *  # noqa: F401, F403
from .report import *  # noqa: F401, F403
from .scheduler import *  # noqa: F401, F403
from .user import *  # noqa: F401, F403



import json
from .base import BaseResource


class JobResource(BaseResource):
    def __init__(self, *args, **kwargs):
        super(JobResource, self).__init__(*args, **kwargs)

    def on_get(self, req, resp):
        '''List all job instances'''
        resp.content_type = 'application/json'
        resp.body = json.dumps(self.db.jobs.list(req.context['user'], req.params, self.session))

    def on_post(self, req, resp):
        '''Create new or delete job instance'''
        resp.content_type = 'application/json'
        action = req.params.get('action')
        if action == 'delete':
            resp.body = json.dumps(self.db.jobs.delete(req.context['user'], req.params, self.session, self.scheduler))
        else:
            resp.body = json.dumps(self.db.jobs.store(req.context['user'], req.params, self.session, self.scheduler))


class JobDetailResource(BaseResource):
    def __init__(self, *args, **kwargs):
        super(JobDetailResource, self).__init__(*args, **kwargs)

    def on_get(self, req, resp):
        '''Get details of specific job instance'''
        resp.content_type = 'application/json'
        resp.body = json.dumps(self.db.jobs.detail(req.context['user'], req.params, self.session))
